from setuptools import setup, find_packages

setup(
    name='myjuliamod',
    version='0.0.1',
    description='myjuliamod example',
    author='John Lapeyre',
    packages=find_packages(),
    install_requires=['julia>=0.2',
                      'julia_project>=0.0.23'
                      ],
    include_package_data=True
)



# Is there any way to enforce this ?

class CallJulia:

    def seval(self, code: str):
        """Evaluate `code` in Julia. Only one toplevel statement allowed"""

    def seval_all(self, code: str):
        """Evaluate `code` in Julia"""

    # These properties are defined
    # self.julia_path = None
    # self.depot_dir = None
    # self.data_path = None
    # self.julia_system_image = None
    # self.questions = None
    # self.libjulia # : lib.LibJulia



# This only creates the JuliaProject instance, but does not install or initialize the Julia environment
# The user must take some action to install and initialize. For example myjuliamod.project.ensure_init()
from ._julia_project import project

# You may choose to uncomment the following, in which case the julia installation is
# checked and initialized every time the user does `import myjuliamod`.
#
# In the files hellomod.py and _julia_project.py we included code to automatically
# load a submodule (hellomod) after initialization, and to import symbols into
# the myjuliamod namespace. Those tricks are not necessary if you uncomment the following.
#
# project.ensure_init()
# import hellomod
# from hellomod import hello
#
# In practice, you probably would not want to import both hellomod and hello.



class LibJulia:
    """Store information on loaded libjulia"""

    def __init__(self,
                 libjulia,
                 libjulia_path,
                 bindir
                 ):
        self.libjulia = libjulia
        self.libjulia_path = libjulia_path
        self.bindir = bindir



import os
import ctypes
import subprocess
import logging

import juliacall
# Comment in original source:
from juliacall import CONFIG

from .calljulia import CallJulia # interface
from . import lib
from . import utils

# This is perhaps heavier than we need. It borrows from pyjulia
# from . import libjulia as libjulia_mod


LOGGER = logging.getLogger('julia_project.juliacall')
LOGGER.info("importing julia_project/juliacall")


def load_libjulia(julia_path,
                  project_path,
                  system_image=None):
    # Find the Julia executable and project
    CONFIG['exepath'] = exepath = julia_path # juliapkg.executable()
    CONFIG['project'] = project = project_path # juliapkg.project()

    # Find the Julia library
    cmd = [exepath, '--project='+project, '--history-file=no',
           '--startup-file=no', '-O0', '--compile=min', '-e',
           'import Libdl; print(abspath(Libdl.dlpath("libjulia")))']
    CONFIG['libpath'] = libjulia_path = subprocess.run(cmd, check=True, capture_output=True, encoding='utf8').stdout
    assert os.path.exists(libjulia_path)
    LOGGER.info(f"libjulia_path: {libjulia_path}")

    current_dir = os.getcwd()

    julia_toplevel = os.path.dirname(os.path.dirname(libjulia_path))
    bindir = os.path.realpath(os.path.join(julia_toplevel, "bin"))

    assert project is not None
    try:
        os.chdir(os.path.dirname(libjulia_path))
        CONFIG['lib'] = libjulia = ctypes.PyDLL(libjulia_path, ctypes.RTLD_GLOBAL) # <-- avoids segfault
        if os.getenv("JULIA_PROJECT") != project:
            print("JULIA_PROJECT NOT SET!")
            os.environ['JULIA_PROJECT'] = project
        LOGGER.info(f"setting JULIA_PROJECT = {project}")
        if system_image is not None and os.path.exists(system_image):
            LOGGER.info(f"jl_init_with_image({bindir.encode('utf8')},{system_image.encode('utf8')}")
            libjulia.jl_init_with_image__threading.argtypes = []
            libjulia.jl_init_with_image__threading.restype = None
            libjulia.jl_init_with_image__threading(
                bindir.encode('utf8'),
                system_image.encode('utf8'))
        else:
            LOGGER.info("jl_init with default system image")
            libjulia.jl_init__threading.argtypes = []
            libjulia.jl_init__threading.restype = None
            libjulia.jl_init__threading()
        LOGGER.info("libjulia inited.")
        libjulia.jl_eval_string.argtypes = [ctypes.c_char_p]
        libjulia.jl_eval_string.restype = ctypes.c_void_p
    finally:
        os.chdir(current_dir)

    return libjulia, libjulia_path, bindir


def init_pythoncall(libjulia, project_path):
    os.environ['JULIA_PYTHONCALL_PROJECT'] = project_path
    if utils.has_manifest_toml(project_path):
        do_resolve = "false"
    else:
        do_resolve = "true"
    script = f'''
             try
                import Pkg
                Pkg.activate(ENV["JULIA_PYTHONCALL_PROJECT"], io=devnull)
                pt = Base.parsed_toml(Pkg.project().path)
                if ! any(==("PythonCall"), keys(pt["deps"]))
                    Pkg.add("PythonCall")
                    Pkg.resolve()
                    Pkg.instantiate()
                elseif {do_resolve}
                    Pkg.resolve()
                    Pkg.instantiate()
                end
                try
                    import PythonCall
                catch
                    Pkg.resolve()
                    Pkg.instantiate()
                    import PythonCall
                end
            catch err
                print(stderr, "ERROR: ")
                showerror(stderr, err, catch_backtrace())
                flush(stderr)
                rethrow()
            end
            '''
    LOGGER.info("Activating project")
    res = libjulia.jl_eval_string(script.encode('utf8'))
    if res is None:
        raise Exception('PythonCall.jl did not start properly')

    CONFIG['inited'] = True
    LOGGER.info("Done juliacall init .....")


class JuliaCall(CallJulia):

    julia = juliacall

    def __init__(self,
                 julia_path,
                 project_path=None,
                 julia_system_image=None,
                 use_sys_image=None,
                 ):
        self.julia_path = julia_path
        self.project_path = project_path
        self.julia_system_image = julia_system_image
        self._is_initialized = False
        self.use_sys_image = use_sys_image
        self.libjulia = None


    # @classmethod
    # def name(cls):
    #     return "juliacall"

    # bug requires stripping.
    # pylint: disable=no-member
    def seval(self, _str):
        return juliacall.Main.seval(_str.strip())


    # pylint: disable=no-member
    def seval_all(self, _str):
        expr = juliacall.Main.Meta.parseall(_str)
        return juliacall.Main.eval(expr)


    def beval(self, _str):
        return self.libjulia.libjulia.jl_eval_string(_str.encode('utf8'))


    def simple_import(self, module : str):
        """
        import the julia module `module` and return the python-wrapped module.
            `Example = self.simple_import("Example")`
        """
        self.seval("import " + module)
        return self.seval(module)


    def start_julia(self):
        if not self._is_initialized:
            if CONFIG['inited']:
                LOGGER.info("CONFIG['inited'] is true, but we are not reall initialized.")
            LOGGER.info("Initializing julia.")
            if self.use_sys_image is not False:
                system_image = self.julia_system_image.sys_image_path
            else:
                system_image = None
            libjulia, libjulia_path, bindir = load_libjulia(
                self.julia_path,
                self.project_path,
                system_image=system_image
            )
            assert isinstance(self.project_path, str)
            init_pythoncall(libjulia, self.project_path)
            self.libjulia = lib.LibJulia(libjulia, libjulia_path, bindir)
            self._is_initialized = True
            CONFIG['inited'] = True



import sys

# Import the instance of JuliaProject
from ._julia_project import project

# The user may have already called this line to initialize the Julia environment.
# Calls to ensure_init after the first call are no-ops.
# Because we include this line, the user may enter `import myjuliamod.hellomod`
# without having first imported myjuliamod or having run ensure_init().
# Omit this line if you want to force the user to call ensure_init explicitly.
project.ensure_init()


# Import a Julia module. This works with either julia/PyCall or juliacall/PythonCall
Example = project.simple_import("Example")

# Alternatively, For julia/PyCall
# from julia import Example

# And for juliacall/PythonCall
# import juliacall
# juliacall.using(locals(), module="Example", prefix="")


def hello():
    return Example.hello("myjuliamod")

# You may want to do something like the following.
# This has the same effect as putting the following line in __init__.py:
# from .hellomod import hello
# Ths difference is that the previous line would start Julia upon importing myjuliamod, which
# may not be desirable.
sys.modules['myjuliamod'].hello = hello



# import julia Surely don't want thisw
import logging
import importlib

from julia_project import JuliaProject

# The top-level directory of the mymodule installation must be
# passed when constructing JuliaProject. We compute this path here.
import os
myjuliamod_path = os.path.dirname(os.path.abspath(__file__))

# You may supply a hook to run after JuliaProject.ensure_init is called.
# For example, here we import hellomod into myjuliamod. The effect is
# as if we had added the line `import .hellomod` to mymjuliamod.__init__.
# We did not do the latter, because that would require checking and
# initializing Julia upon importing myjuliamod. The package author
# (i.e. author of myjuliamod) may wish to avoid the latter.
def _post_init_hook():
    importlib.import_module('.hellomod', 'myjuliamod')


project = JuliaProject(
    name="myjuliamod",
    package_path=myjuliamod_path,
    version_spec = "^1.6", # Must be at least 1.6
    env_prefix = 'MYJULIAMOD_', # env variables prefixed with this may control JuliaProject
    logging_level = logging.INFO, # or logging.WARN,
    console_logging=False,
    post_init_hook=_post_init_hook, # Run this after ensure_init
#    registries = {"Bogus" : "http://badurlbogus.com"}
#    calljulia_lib = "pyjulia"
    calljulia = "juliacall"
)



import os
import logging
from . import utils
import julia_project_basic as basic
import julia as pyjulia_julia

LOGGER = logging.getLogger('julia_project.system_image') # shorten this?


class JuliaSystemImage:
    """
    This class manages compilation of a Julia system image.
    """
    def __init__(self,
                 name,
                 sys_image_dir, # Absolute, not relative path!
                 julia_path,
                 project_path,
                 julia_version = None,
                 sys_image_file_base=None,
                 ):
        self.sys_image_dir = sys_image_dir
        if sys_image_file_base is None:
            sys_image_file_base = "sys_" + name
        self.sys_image_file_base = sys_image_file_base
        self.julia_version = julia_version
        # self.sys_image_path is the path for the system image written upon compilation. This
        # file will be renamed after compilation.
        self.sys_image_path = self._in_sys_image_dir(
            self.sys_image_file_base + "-" + self.julia_version + utils.SHLIB_SUFFIX
        )
        self.compiled_system_image = self._in_sys_image_dir("sys_julia_project" + utils.SHLIB_SUFFIX)
        self.calljulia = None # Can't pass this. It must be set later.
        self.julia_path = julia_path
        self.project_path = project_path


    # This must be set after __init__, because calljulia is instantiated with data from self
    def set_calljulia(self, calljulia):
        self.calljulia = calljulia


    def _in_sys_image_dir(self, rel_path):
        """Return absolute path from path relative to system image build dir."""
        return os.path.join(self.sys_image_dir, rel_path)


    def clean(self):
        """
        Delete some files created when installing Julia packages. These are Manifest.toml files
        and a compiled system image.
        """
        for _file in [self._in_sys_image_dir("Manifest.toml"), self._in_sys_image_dir("JuliaManifest.toml"),
                      self.sys_image_path]:
            utils.maybe_remove(_file)


    def compile(self):
        """
        Compile a system image for the dependent Julia packages in the subdirectory `./sys_image/`. This
        system image will be loaded the next time you import the Python module.
        """
        Main = self.calljulia.julia.Main
        Pkg = self.calljulia.julia.Pkg
        current_path = Main.pwd()
        current_project = Pkg.project().path
        old_julia_project = os.getenv("JULIA_PROJECT")
        try:
            self._compile()
        except:
            print("Exception when compiling system image.")
            raise
        finally:
            basic.reset_env_var("JULIA_PROJECT", old_julia_project)
            Main.cd(current_path)
            Pkg.activate(current_project)


    def _ensure_pycall_pythoncall_imported(self):
        Pkg = self.calljulia.julia.Pkg
        project_path = self.project_path
        deps = basic.parse_project(project_path)["deps"].keys()
        depot_path = self.calljulia.julia.Main.DEPOT_PATH[0]
        pycall_ok = basic.test_pycall(
            project_path, self.julia_path, depot_path=depot_path
        )["pycall_ok"]
        if not "PyCall" in deps and pyjulia_julia.find_libpython.linked_libpython() is not None and pycall_ok:
            Pkg.add("PyCall")
            self.calljulia.simple_import("PyCall")
        if not "PythonCall" in deps:
            Pkg.add("PythonCall")
        import juliacall  # Otherwise import below fails
        self.calljulia.simple_import("PythonCall")


    def _compile(self):
        """
        Compile a Julia system image with all requirements for the julia project.
        """
        Main = self.calljulia.julia.Main
        Pkg = self.calljulia.julia.Pkg
        if not os.path.isdir(self.sys_image_dir):
            msg = f"Can't find directory for compiling system image: {self.sys_image_dir}"
            raise FileNotFoundError(msg)

        if not utils.has_project_toml(self.sys_image_dir):
            msg = utils.no_project_toml_message(self.sys_image_dir)
            LOGGER.error(msg)
            raise FileNotFoundError(msg)

        self._ensure_pycall_pythoncall_imported()

        for _file in [self._in_sys_image_dir("Manifest.toml"), self._in_sys_image_dir("JuliaManifest.toml")]:
            utils.maybe_remove(_file)
        Main.eval('ENV["PYCALL_JL_RUNTIME_PYTHON"] = Sys.which("python")')
        assert isinstance(self.sys_image_dir, str)
        Pkg.activate(self.sys_image_dir)
        os.environ["JULIA_PROJECT"] = self.sys_image_dir
        pycall_loaded = Main.is_loaded("PyCall")
        pythoncall_loaded = Main.is_loaded("PythonCall")
        deps = basic.parse_project(self.sys_image_dir)["deps"].keys() # This is faster
#        deps = Main.parse_project()["deps"].keys()
        pycall_in_deps = "PyCall" in deps
        pythoncall_in_deps = "PythonCall" in deps
        # Let's try always including both of these. This might prevent crashes when
        # creating sys image with one of them, and loading it with the other
        if pycall_loaded and not pycall_in_deps:
            Pkg.add("PyCall")
            self.calljulia.simple_import("PyCall")
        if not pythoncall_in_deps:
            Pkg.add("PythonCall")
        import juliacall
        self.calljulia.simple_import("PythonCall")
        LOGGER.info("Compiling: probed Project.toml path: %s", Pkg.project().path)
        Main.cd(self.sys_image_dir)
        try:
            Pkg.resolve()
        # Assume that failure of resolve is because update() has not been called
        # TODO: Find more precisely what error is raised.
        except Exception:
            msg = "Pkg.resolve() failed. Updating packages."
            LOGGER.info(msg)
            Pkg.update()
            Pkg.resolve()
        Pkg.instantiate()

        # Following will also perform compilation, with more granual error messages
        # But, it is harder to read.
        # cj = self.calljulia
        # PackageCompiler = cj.simple_import("PackageCompiler")
        # Libdl = cj.simple_import("Libdl")
        # cj.seval_all("""
        #    ENV["PYCALL_JL_RUNTIME_PYTHON"] = Sys.which("python")
        #    ENV["PYTHON"] = Sys.which("python")
        # """)
        # Main.include("packages.jl")
        # if pycall_loaded:
        #     LOGGER.info("push!(packages, :PyCall)")
        #     cj.seval_all("push!(packages, :PyCall)")
        # if pythoncall_loaded:
        #     LOGGER.info("push!(packages, :PythonCall)")
        #     cj.seval_all("push!(packages, :PythonCall)")
        # cj.seval('sysimage_path = joinpath(@__DIR__, "sys_julia_project." * Libdl.dlext)')
        # cj.seval_all("""PackageCompiler.create_sysimage(packages; sysimage_path=sysimage_path,
        # precompile_execution_file=joinpath(@__DIR__, "compile_exercise_script.jl"))""")

        packages_file = os.path.join(self._in_sys_image_dir("packages.jl"))
        if not os.path.exists(packages_file):
            raise FileNotFoundError(f'{packages_file} does not exist')

        _bool = {True: "true", False: "false"}

        cscript = f'''
        import PackageCompiler
        using Libdl: Libdl
        let
          ENV["PYCALL_JL_RUNTIME_PYTHON"] = Sys.which("python")
          ENV["PYTHON"] = Sys.which("python")
          packages = include("packages.jl")
          if {_bool[pycall_loaded]}
             push!(packages, :PyCall)
          end
          if {_bool[pythoncall_loaded]}
             push!(packages, :PythonCall)
          end
          sysimage_path = joinpath(@__DIR__, "sys_julia_project." * Libdl.dlext)

          if isfile("compile_exercise_script.jl")
            PackageCompiler.create_sysimage(packages; sysimage_path=sysimage_path,
                precompile_execution_file=joinpath(@__DIR__, "compile_exercise_script.jl"),
                incremental=true,
                project=joinpath(@__DIR__, "."))
          else
            PackageCompiler.create_sysimage(packages; sysimage_path=sysimage_path,
                incremental=true,
                project=joinpath(@__DIR__, "."))
          end
        end
        '''
        LOGGER.info("Running compile script.")
        self.calljulia.seval_all(cscript)
        if os.path.isfile(self.compiled_system_image):
            LOGGER.info("Compiled image found: %s.", self.compiled_system_image)
            os.rename(self.compiled_system_image, self.sys_image_path)
            LOGGER.info("Renamed compiled image to: %s.", self.sys_image_path)
            if not os.path.isfile(self.sys_image_path):
                LOGGER.error("Failed renamed compiled image to: %s.", self.sys_image_path)
                raise FileNotFoundError(self.compiled_system_image)
        else:
            raise FileNotFoundError(self.compiled_system_image)



__version__ = "0.1.27"

import os
import sys
import ctypes
import shutil
from ._julia_project import JuliaProject

os.environ["PYTHON_JULIACALL_NOINIT"] = "yes"
os.environ["JULIA_PYTHONCALL_EXE"] = sys.executable or ''
os.environ['JULIA_PYTHONCALL_LIBPTR'] = str(ctypes.pythonapi._handle) or ''
os.environ['PYTHON'] = shutil.which("python") or ''

# This is only for debugging
# import julia
# import julia.core
# julia.core.enable_debug()



from . import utils
from . import environment

_QUESTIONS = {'install' : "No Julia installation found. Would you like jill.py to download and install Julia?",
              'compile' :
"""
I can compile a system image after installation.
Compilation may take a few, or many, minutes. You may compile now, later, or never.
Would you like to compile a system image after installation?
""",
              'depot' :
"""
You can install all of the Julia packages and package information in a
module-specific "depot", that is, one specific to this Python module. This may
allow you to use Julia with python projects that have different Python
installation locations.  If you answer "no", packages will be installed in the
standard per-user Julia "depot".

Would you like to use a python-module-specific depot for Julia packages?
"""
              }


class ProjectQuestions:

    def __init__(self,
                 depot=None,
                 env_vars=None,
                 logger=None,
                 ):
        self.results = {"install": None, "compile": None, "depot": depot}
        if env_vars is None:
            env_vars = environment.EnvVars()
        self.logger = logger
        self._env_vars = env_vars


    def ask_question(self, question_key):
        if self.results[question_key] is None:
            result = utils.query_yes_no(_QUESTIONS[question_key])
            if result is not None:
                self.results[question_key] = result
                self.logger.info(f"Question '{question_key}', answered {result}")
                return result
            else:
                self.logger.info(f"Question '{question_key}', not answered. Possibly interrupted.")
                return None


    def ask_questions(self):
        result = None
        for q in self.results:
            result = self.ask_question(q)
        return result


    def _read_one_variable(self, var_base_name, question_key):
        var_name = self._env_vars.envname(var_base_name)
        result = self._env_vars.getenv(var_base_name)
        if result:
            if result == 'y':
                self.results[question_key] = True
                self.logger.info(f"read {var_name} = 'y'")
            elif result == 'n':
                self.results[question_key] = False
                self.logger.info(f"read {var_name} = 'n'")
            else:
                raise ValueError(f"{var_name} must be y or n")
        else:
            self.logger.info(f"{var_name} not set")


    def read_environment_variables(self):
        self._read_one_variable("INSTALL_JULIA", "install")
        self._read_one_variable("COMPILE", "compile")
        self._read_one_variable("DEPOT", "depot")



import os

class EnvVars:

    def __init__(self,
                 env_prefix="JULIA_PROJECT_",
                 ):
        self.env_prefix = env_prefix


    def envname(self, env_var):
        return self.env_prefix + env_var


    def getenv(self, env_var):
        return os.getenv(self.env_prefix + env_var)



import logging
import os
import shutil
import warnings
import distutils.dir_util
import find_julia
import julia.find_libpython
from .julia_system_image import JuliaSystemImage
from . import utils
import julia_project_basic as basic

from .environment import EnvVars
from .questions import ProjectQuestions


# The depot referred to here is the default depot, i.e. ~/.julia
# This true even if the user requests a "private" depo.
# In the latter case, we will have a depot within the default depot.
def _get_parent_project_path():
    env_path = utils.get_virtual_env_path()
    if env_path is None:
        env_path = utils.default_depot_path()
    return os.path.join(env_path, "julia_project")


def _calljulia_lib(calljulia : str, logger=None):
    if calljulia == "pyjulia":
        if logger:
            logger.info("importing PyJulia")
        from .pyjulia import PyJulia
        return PyJulia
    if calljulia == "juliacall":
        if logger:
            logger.info("importing JuliaCall")
        from .juliacall import JuliaCall
        return JuliaCall
    raise ValueError(f"calljulia must be 'pyjulia' or 'juliacall'. Got {calljulia}")


def _validate_calljulia(calljulia : str):
    if calljulia not in ["pyjulia", "juliacall", None]:
        raise ValueError(f'calljulia must be one of "pyjulia", "juliacall", `None`. Got {calljulia}')


class JuliaProject:
    """
    A class to manage a Julia project with a Python front end.

    The Julia project exists within a python module.

    This class manages
    1) Setting up the julia module provided by pyjulia, including building PyCall.jl
    2) Downloading and installing Julia packages (and a registry)
    3) Setting up the Julia module, including loading a custom system image
    4) Loading Julia modules
    5) Compiling a system image for the Julia Project

    Parameters
    ----------
    name : str
        The name of the project. Used to construct system image file name and in the logger.
    package_path : str
        The top level directory of the installed Python project that is using julia_project
    registries : dict, optional
        A dict whose keys are registry names and values are urls of Julia registries.
        These registries will be installed automatically.
    version_spec : str or object, optional
        A Julia version specification that the Julia executable must satisfy. Default "^1".
    strict_version : bool  If `True` disallow prerelease (development) versions of Julia. Default `True`.
    sys_image_dir : str
        The directory under `package_path` in which the system image is built. Defaults to "system_image".
    sys_image_file_base : str
        The basename for the system image file. The julia version is appended to this name. Defaults to
        `"sys_" + name`.
    env_prefix : str
        A prefix to all environment variables controlling the project. Defaults to "JULIA_PROJECT_".
    depot : bool, optional
        Whether a project-specific depot will be used, rather than the default depot. If `None`, then
        the user may be prompted for a value.
    post_init_hook : function
         A function to call after the init method.
    logging_level
        The logging level. For example `logging.INFO`. Optional. If omitted, then no logging is done.
    console_logging : bool
        If `True` then log to the console as well as to a file.
    """

    def __init__(self,
                 name,
                 package_path,
                 registries=None,
                 version_spec=None,
                 strict_version=True,
                 sys_image_dir="sys_image",
                 sys_image_file_base=None,
                 env_prefix="JULIA_PROJECT_",
                 depot=None,
                 post_init_hook=None,
                 pre_instantiate_cmds=None,
                 logging_level=None,
                 console_logging=False,
                 calljulia = "pyjulia",
                 ):

        self.name = name
        self.package_path = package_path
        self.julia_path = None
        self.registries = registries
        self.rel_sys_image_dir = sys_image_dir
        self.sys_image_file_base = sys_image_file_base # May be None
        self._env_vars = EnvVars(env_prefix)
        self._logging_level = logging_level
        self._console_logging = console_logging
        self.questions = ProjectQuestions(depot=depot, env_vars=self._env_vars)
        self._init_flags = {"initialized": False, "initializing": False, "disabled": False}
        self._post_init_hook = post_init_hook
        self._pre_instantiate_cmds = pre_instantiate_cmds
        os.environ['PYCALL_JL_RUNTIME_PYTHON'] = shutil.which("python") or ''
        self.version_spec = version_spec
        self.strict_version = strict_version
        if calljulia is None:
            calljulia = "pyjulia"
        _validate_calljulia(calljulia)
        self._calljulia_name = calljulia
        self._use_sys_image = None
        self.logger = None # TODO: Use global LOGGER
        self.log_file_path = None # TODO: can we git rid of this?
        self.julia = None # TODO: We need a better name for this. Its value is either julia or juliacall
        self.calljulia = None
        self.julia_system_image = None
        self.sys_image_dir = None
        self.julia_version = None
        self.project_path = None
        self.depot_path = None


    def _in_package_dir(self, rel_path):
        """Return absolute path from path relative to installation dir."""
        return os.path.join(self.package_path, rel_path)


    def _in_project_dir(self, rel_path):
        return os.path.join(self.project_path, rel_path)


    def _set_project_path(self):
        self.project_path = os.path.join(_get_parent_project_path(), self.name + "-" + self.julia_version)
        os.makedirs(self.project_path, exist_ok = True)
        assert self.project_path is not None
        os.environ["JULIA_PROJECT"] = self.project_path
        self.logger.info(f'os.environ["JULIA_PROJECT"] = {self.project_path}')
        utils.update_copy(self._in_package_dir("Project.toml"), self._in_project_dir("Project.toml"))
        utils.update_copy(self._in_package_dir("JuliaProject.toml"), self._in_project_dir("JuliaProject.toml"))
        if not utils.has_project_toml(self.project_path):
            msg = utils.no_project_toml_message(self.project_path)
            self.logger.error(msg)
            raise FileNotFoundError(msg)


    def disable_init(self):
        """Make calling `ensure_init` do nothing.
        Returns `True` if `ensure_init` was already disabled before calling `disable_init`.
        Otherwise returns `False`. By default, `ensure_init` is enabled.
        """
        if self.is_initialized:
            raise RuntimeError(
                "This JuliaProject is already initialized. Disabling initialization makes no sense."
            )
        old_val = self._init_flags['disabled']
        self._init_flags['disabled'] = True
        return old_val


    def enable_init(self):
        """Enable normal operation of calls to `ensure_init`.
        Returns `True` if `ensure_init` was disabled before calling `enable_init`.
        Otherwise returns `False`. By default, `ensure_init` is enabled.
        """
        if self.is_initialized:
            raise RuntimeError(
                "This JuliaProject is already initialized. Enabling initialization makes no sense."
            )
        old_val = self._init_flags['disabled']
        self._init_flags['disabled'] = False
        return old_val


    @property
    def is_initialized(self):
        """This property has value `True` if `ensure_init` has been called and no errors were detected."""
        return self._init_flags['initialized']


    def ensure_init(self,
                    calljulia=None,
                    depot=None,
                    use_sys_image=None,
                    compile=None,
                    install_julia=None,
                    julia_path=None,
                    version_spec=None,
                    strict_version=None,
                    pre_instantiate_cmds=None
                    ):
        """
        Initializes the Julia project if it has not yet been initialized.

        If any optional parameter is set to a value other than `None`, then it overrides the
        setting of the corresponding parameter elsewhere, in particular in the creation of the
        `JuliaProject` instance. If not specified here, they may be set automatically, or by
        environment variables, or by asking questions.  Initialization includes searching for the
        Julia executable, optionally installing it, if not found. Resolving package requirements
        and downloading and installing them. Optionally compiling a system image.

        Subequent calls to `ensure_init` do nothing. The exception is if you call `ensure_init`
        again with a different value of `calljulia`, which will raise a `ValueError`.

        Args:
            calljulia : str The interface library to use, either "pyjuia" or "juliacall". Default "pyjulia"

            depot : bool If True, install a Julia depot (where packages are stored) in the data
                directory for this project. If False or None, use the default, common, Julia depot.
                Using a project-specific depot avoids the possibility that PyCall.jl will be rebuilt
                frequently if it is used in different projects.

            use_sys_image : bool If `False`, then don't load a custom system image, even if it is present.

            compile : bool Whether to compile a system image after initialization.

            julia_path : str The path to a Julia executable.

            version_spec : str A (Julia compat) version specification used to filter julia executables.

            install_julia : bool Whether to install julia if no executable is found.

            pre_instantiate_cmds : str a string of Julia commands that will be executed immediately before
                   instantating the project. `Pkg` will be imported before they are executed.

            strict_version : bool If `True` then pre-release versions will be excluded when searching for
                the Julia exectuable.
        """
        if not self._init_flags['initialized'] and not self._init_flags['disabled'] and not self._init_flags['initializing']:
            # if self._init_flags['initializing']:
            #     print("Initialization was aborted or failed. Trying again.")
            if use_sys_image is not None:
                self._use_sys_image = use_sys_image
            _validate_calljulia(calljulia)
            if calljulia is not None:
                self._calljulia_name = calljulia
            if (self._calljulia_name == "pyjulia" and
                julia.find_libpython.linked_libpython() is None):
                warnings.warn(
"""
Your python executable is statically linked to libpython and you (or a package
author) have requested the Python module `julia` from the package `pyjulia`. The
module `julia` requires a dynamically linked python executable. The module
`juliacall` from the package `juliacall` will be used instead. This may not be
compatible with package that created this instance of JuliaProject.
""")
                self._calljulia_name = "juliacall"
            if strict_version is not None:
                self.strict_version = strict_version
            if depot is not None:
                self.questions.results['depot'] = depot
            if install_julia is not None:
                self.questions.results['install'] = install_julia
            if julia_path is not None:
                self.julia_path = os.path.expanduser(julia_path)
                self.questions.results['install'] = False
            if compile is not None:
                self.questions.results['compile'] = compile
            if version_spec is not None:
                self.version_spec = version_spec
            self._pre_instantiate_cmds = pre_instantiate_cmds

            try:
                self._init_flags['initializing'] = True
                self.init()
            except:
                print("Initialization failed. You may try running again")
                raise
            finally:
                self._init_flags['initializing'] = False
        # Reiniting is a no-op
        elif self._init_flags['initialized'] and calljulia is not None:
            incompat_reinit = ((self.julia.__name__ == 'julia' and calljulia != 'pyjulia')
                               or
                               (self.julia.__name__ == 'juliacall' and calljulia != 'juliacall')
                               )
            if incompat_reinit:
                raise ValueError(f"Can't change library. Already initialzed with '{self.julia.__name__}'")


    def init(self):
        """Run all steps to initialize and load Julia and the Julia project."""
        self._setup_logging()
        self.logger.info("")
        self.logger.info("JuliaProject.init()")
        self.questions.logger = self.logger
        self.questions.read_environment_variables()
        if self.julia_path is None:
            self._find_julia()
        if self.julia_path is None:
            raise FileNotFoundError("No julia executable found")
        self.logger.info(f"julia path: {self.julia_path}")
        self.julia_version = utils.julia_version_str(self.julia_path)
        self.logger.info("Julia version: %s.", self.julia_version)
        self._set_project_path()
        self.sys_image_dir = self._in_project_dir(self.rel_sys_image_dir)
        package_sys_image_dir = os.path.join(self.package_path, self.rel_sys_image_dir)
        if os.path.exists(package_sys_image_dir):
            self.logger.info(f"Copying/updating installed system image directory {self.sys_image_dir}")
            distutils.dir_util.copy_tree(
                package_sys_image_dir, self.sys_image_dir, update=1)
        else:
            self.logger.info(f"System image dir source not found at {package_sys_image_dir}")
        possible_depot_path = self._in_project_dir("depot")
        if os.path.isdir(possible_depot_path) and not self.questions.results['depot'] is False:
            self.questions.results['depot'] = True
            self.logger.info("Found existing Python-project specific Julia depot")
        if self.questions.results['depot'] is None and self._calljulia_name == "juliacall":
            # Only PyCall needs the possibility of a special depot
            # Not clear whether to ask in the case of juliacall or not
            self.questions.results['depot'] = False

        if self.questions.results['depot'] is True:
            self.depot_path = possible_depot_path
        else:
            self.depot_path = None

        self.julia_system_image = JuliaSystemImage(
            self.name,
            sys_image_dir=self.sys_image_dir,
            julia_path = self.julia_path,
            project_path = self.project_path,
            sys_image_file_base=self.sys_image_file_base,
            julia_version = self.julia_version,
            )
        calljulia_lib = _calljulia_lib(self._calljulia_name, self.logger)



        if self._calljulia_name == "pyjulia":
            needed_packages = ["PyCall"]
        elif self._calljulia_name == "juliacall":
            needed_packages = ["PythonCall"]
            # We do this above. So comment it out and test.
            # Only PyCall needs the possibility of a special depot
            # if self.questions.results['depot'] is None:
            #     self.questions.results['depot'] = False
        else:
            needed_packages = None


        _need_resolve = basic.need_resolve(self.project_path, self.depot_path)
        _packages_to_add = basic.packages_to_add(self.project_path, needed_packages)
        if _need_resolve or _packages_to_add:
            self.questions.ask_questions()
            if self.questions.results['depot'] is True: # May have changed depot
                self.depot_path = possible_depot_path
            else:
                self.depot_path = None

        def answer_rebuild_callback():
            self.questions.results['depot'] = False # only one or the other
            self.questions.ask_questions()

        def answer_depot_callback():
            self.questions.results['depot'] = True
            self.questions.ask_questions()


        # print(f"Questions are {self.questions.results}")
        # ensure that packages, registries, etc. are installed
        if self._calljulia_name != "pyjulia":
            basic.ensure_project_ready(
                project_path=self.project_path,
                julia_exe=self.julia_path,
                depot_path=self.depot_path,
                registries=self.registries,
                needed_packages=needed_packages,
                pre_instantiate_cmds=self._pre_instantiate_cmds,
                clog=True,
                pre_install_callback=None # we now do this above self.questions.ask_questions,
            )
        else:
            basic.ensure_project_ready_fix_pycall(
                project_path=self.project_path,
                julia_exe=self.julia_path,
                depot_path=self.depot_path,
                possible_depot_path=possible_depot_path,
                registries=self.registries,
                needed_packages=needed_packages,
                pre_instantiate_cmds=self._pre_instantiate_cmds,
                clog=True,
                pre_install_callback=self.questions.ask_questions,
                question_callback=None, # self.questions.deal_with_incompatibility,
                answer_rebuild_callback=answer_rebuild_callback,
                answer_depot_callback=answer_depot_callback
            )

        if self.questions.results['depot'] is True:
            assert possible_depot_path is not None
            os.environ["JULIA_DEPOT_PATH"] = possible_depot_path
            self.logger.info(f"Using private depot '{possible_depot_path}'")
        else:
            self.logger.info("Using default depot.")

        self.calljulia = calljulia_lib(
            self.julia_path,
            project_path=self.project_path,
            julia_system_image=self.julia_system_image,
            use_sys_image=self._use_sys_image,
        )
        # Ugh. Don't really like doing it this way.
        self.julia_system_image.set_calljulia(self.calljulia)
        # Start the Julia runtime via libjulia
        self.calljulia.start_julia()
        self.julia = self.calljulia.julia # More convenient to get at this level
        self._load_julia_utils()
        # Either `julia` or `juliacall`: The Python/Julia interface module.
        self.julia = self.calljulia.julia
        # Redundant
        # pylint: disable=no-member
        self.logger.info(f'PyCall version: {self.julia.Main.pycall_version()}')
        self.logger.info(f'PythonCall version: {self.julia.Main.pythoncall_version()}')
        if self._post_init_hook is not None:
            self._post_init_hook()
        self._init_flags['initialized'] = True
        # Note that we consider initialization to have succeeded before we run the compilation
        # But, the post_init_hook is part of initialization
        if self.questions.results['compile']:
            self.compile()


    @property
    def using_pyjulia(self):
        """Has value `True` only if this instance of `JuliaProject` is using or will use `pyjulia` (the
        Python module `julia`).

        Under some circumstances this may change between between creating the `JuliaProject`
        instance and calling `ensure_init`.
        """
        assert self._calljulia_name
        return self._calljulia_name == "pyjulia"


    @property
    def using_juliacall(self):
        """Has value `True` only if this instance of `JuliaProject` is using or will use `juliacall`.

        Under some circumstances this may change between between creating the `JuliaProject`
        instance and calling `ensure_init`.
        """
        assert self._calljulia_name
        return self._calljulia_name == "juliacall"


    # For testing. Build PyCall with "wrong" libpython
    def _build_pycall_conda(self):
        basic.rebuild_pycall(
            self.project_path,
            python_exe="conda",
            julia_exe=self.julia_path,
            clog=True,
        )


    def _load_julia_utils(self):
        srcdir = os.path.dirname(os.path.realpath(__file__))
        utilf = os.path.join(srcdir, "utils.jl")
        self.calljulia.seval(f'include("{utilf}")')


    def _setup_logging(self):
        self.logger = logging.getLogger('julia_project')
        if self._logging_level is None:
             # fh = logging.NullHandler() # probably don't need this
            return None

        result = self._env_vars.getenv("LOG_PATH")
        if result:
            self.log_file_path = result
        else:
            self.log_file_path = self.name + '.log'
        fh = logging.FileHandler(self.log_file_path)
        logging_level = self._logging_level
        self.logger.setLevel(logging_level)
        fh.setLevel(logging_level)
#        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)

        if self._console_logging:
            ch = logging.StreamHandler()
            ch.setLevel(self._logging_level)
            ch.setFormatter(formatter)
            self.logger.addHandler(ch)
        return None


    def _find_julia(self):
        def other_questions(): # if one question asked, ask all questions at once
            self.questions.ask_question('compile')
            self.questions.ask_question('depot')

        found_path = find_julia.find_or_install(
            env_var = self._env_vars.envname("JULIA_PATH"),
            answer_yes = (self.questions.results['install'] is True),
            version_spec = (self.version_spec if self.version_spec is not None else "^1"),
            post_question_hook = other_questions,
            strict=(self.strict_version if self.strict_version is not None else True)
            )
        if not found_path:
            self.logger.error("No julia executable found or installed.")
            raise FileNotFoundError("No julia executable found or installed.")
        self.julia_path = found_path
        self.questions.results['install'] = False # prevent from being asked again


    # If we use juliacall.using, the module is not imported into Main.
    def simple_import(self, module : str):
        """
        import the julia module `module` and return the python-wrapped module.
        This works with both `pyjulia` and `juliacall`.

        `Example = self.simple_import("Example")`
        """
        self.ensure_init()
        return self.calljulia.simple_import(module)


    def activate_project(self):
        if not utils.has_project_toml(self.project_path):
            msg = utils.no_project_toml_message(self.project_path)
            self.logger.error(msg)
            raise FileNotFoundError(msg)
        Pkg = self.simple_import("Pkg")
        # Activate the Julia project
        Pkg.activate(self.project_path)
        self.logger.info("Probed Project.toml path: %s", Pkg.project().path)


    @property
    def loaded_sys_image(self):
        """The path to the Julia system image as reported by the running libjulia."""
        if not self.is_initialized:
            raise AttributeError(
                "loaded_sys_image is not defined. The project has not been initialized."
            )
        return self.calljulia.seval('unsafe_string(Base.JLOptions().image_file)')


    @property
    def using_custom_sys_image(self):
        """Takes value `True` if a custom system image is loaded.

        If the stock (or unknown) system image is loaded, takes value `False`. Otherwise raise
        an error on accessing.
        """
        loaded = self.loaded_sys_image
        if self.julia_system_image is None:
            raise AttributeError(
                "JuliaSystemImage is not yet created, but system is initialized. Please file a bug report"
            )
        loaded_dir = os.path.dirname(loaded)
        return loaded_dir == self.julia_system_image.sys_image_dir


    def compile(self):
        """
        Compile a system image for the dependent Julia packages in the subdirectory `./sys_image/`. This
        system image will be loaded the next time you import the Python module.
        """
        self.ensure_init()
        if self.using_custom_sys_image:
            warnings.warn("""
You are requesting compiling a custom system image while running a custom-compiled system
image, but this is not safe and not allowed. If you really want to compile a new system
image, please restart your project, passing the keyword argument "use_system_image=False"
to the method `project.ensure_init()`. Then call the method `project.compile()` again.
Alternativley, you can call the method `project.clean()` and restart.
""")
        else:
            self.julia_system_image.compile()


    def clean(self):
        """
        Delete some files created when installing Julia packages. These are Manifest.toml files
        and a compiled system image.
        """
        self.ensure_init()
        for _file in [self._in_project_dir("Manifest.toml"), self._in_project_dir("JuliaManifest.toml"),
                      self._in_package_dir(self.name + '.log')]:
            utils.maybe_remove(_file)
        self.julia_system_image.clean()


    def clean_all(self):
        """
        Delete the (working) Julia project directory. The next time you load the module, a new project
        directory will be created and initialized.
        """
        self.ensure_init()
        project_dir = os.path.join(_get_parent_project_path(), self.name + "-" + self.julia_version)
        if project_dir.find("julia_project") < 0:
            raise ValueError("Expecting project path to contain string 'julia_project'")
        if os.path.isdir(project_dir):
            self.logger.info(f"Removing project directory {project_dir}")
            shutil.rmtree(project_dir)


    def update(self):
        """
        Update the Julia packages that the Python module depends on.

        First, remove possible stale Manifest.toml files and the compiled system image.
        Then update Julia packages and rebuild Manifest.toml file.
        This may help to work around some errors, for instance when compiling.
        """
        self.clean()
        Pkg = self.simple_import("Pkg")
        self.activate_project()
        self.logger.info("Updating Julia packages")
        Pkg.update()
        Pkg.resolve()
        Pkg.instantiate()



import os
import importlib
import logging

import julia
from julia import JuliaError
import julia.core
import julia.api

from .calljulia import CallJulia
from . import lib

LOGGER = logging.getLogger('julia_project.pyjulia')



class PyJulia(CallJulia):

    julia = julia

    def __init__(self,
                 julia_path,
                 project_path=None,
                 julia_system_image=None,
                 use_sys_image=None,
                 ):
        self.julia_path = julia_path
        self.project_path = project_path
        self.julia_system_image = julia_system_image
        self.use_sys_image = use_sys_image
        self.api = None
        self.info = None
        self.libjulia = None


    # @classmethod
    # def name(cls):
    #     return "pyjulia"


    # pylint: disable=no-member
    @classmethod
    def seval(cls, _str):
        return julia.Main.eval(_str.strip())


    # This seems to work with multiple top-level expressions
    # pylint: disable=no-member
    @classmethod
    def seval_all(cls, _str):
        return julia.Main.eval(_str.strip())


    # For before all PyCall/pyjulia code is loaded
    def _seval(self, cmd):
        self.api.jl_eval_string(bytes(cmd.encode('utf8')))


    @classmethod
    def simple_import(cls, module : str):
        """
        import the julia module `module` and return the python-wrapped module.

        `Example = self.simple_import("Example")`
        """
        return importlib.import_module("julia." + module)


    def _load_JuliaInfo(self):
        _info = julia.api.JuliaInfo.load(julia=self.julia_path)
        LOGGER.info("Loaded JuliaInfo.")
        return _info


    # Note, that this doesn't start the Julia runtime. It does everything to prepare for starting.
    def init_julia_module(self):
        info = self._load_JuliaInfo()
        is_compatible_python = info.is_compatible_python()
        is_pycall_built = info.is_pycall_built()
        LOGGER.info("is_pycall_built = %r", is_pycall_built)
        LOGGER.info("is_compatible_python = %r", is_compatible_python)
        # Both or the following should be prevented by install.py. Except if libpython statically linked.
        if not is_pycall_built:
            raise Exception("PyCall is not built.")
        if not is_compatible_python:
            raise julia.core.UnsupportedPythonError(info)

        api = julia.api.LibJulia.from_juliainfo(info)
        LOGGER.info("Loaded LibJulia.")
        self.api = api
        self.info = info
        self.libjulia = lib.LibJulia(api.libjulia, api.libjulia_path, api.bindir)


    def start_julia(self):
        if os.getenv("JULIA_PROJECT") != self.project_path:
            print("JULIA_PROJECT NOT SET!")
        self.init_julia_module()
        sys_image_path = self.julia_system_image.sys_image_path
        if os.path.exists(sys_image_path):
            if self.use_sys_image is not False:
                self.api.sysimage = sys_image_path
                LOGGER.info("Loading system image %s", sys_image_path)
                # pylint: disable=unused-import,import-outside-toplevel
                try:
                    import juliacall
                    # Unfortunately, PythonCall does a lot of work just by existing in the system image.
                    # So this will slow startup time significantly if PythonCall is loaded.
                    # But, because we have PYTHON_JULIACALL_NOINIT = yes, if PythonCall is *not* in
                    # the system image, import juliacall is very fast.
                    LOGGER.info("Loading juliacall to avoid segfault in case PythonCall is in sysimage.")
                except ModuleNotFoundError:
                    pass
            else:
                LOGGER.info("Custom system image found, but will not be used")
        else:
            LOGGER.info(f"No custom system image found: {sys_image_path}.")

        # Both the path and possibly the sysimage have been set. Now initialize Julia.
        LOGGER.info("Initializing julia")
        # These imports import the Julia modules, which are then available outside of this scope
        # pylint: disable=no-member,no-name-in-module,redefined-outer-name,import-error,unused-import,import-outside-toplevel
        try:
            self.api.init_julia()
            LOGGER.info("api.init_julia() done")
            import julia.Base
            import julia.Main
            import julia.Pkg
            LOGGER.info("PyCall, Base, and Main imported")
        except JuliaError as err:
            print("An error occured when initializing Julia.")
            raise err



import ctypes
import shutil
import os
import subprocess

def load_libjulia(julia_path=None, project=None):
    # Find the Julia library
    if julia_path is None:
        julia_path = shutil.which("julia") or ''
    if project is None:
        project =  "."
    cmd = [julia_path, '--project='+project, '--history-file=no',
           '--startup-file=no', '-O0', '--compile=min', '-e',
           'import Libdl; print(abspath(Libdl.dlpath("libjulia")))']
    libjulia_path = subprocess.run(cmd, check=True, capture_output=True, encoding='utf8').stdout
    assert os.path.exists(libjulia_path)

    current_dir = os.getcwd()

    julia_toplevel = os.path.dirname(os.path.dirname(libjulia_path))
    bindir = os.path.realpath(os.path.join(julia_toplevel, "bin"))

    try:
        os.chdir(os.path.dirname(libjulia_path))
        libjulia = ctypes.PyDLL(libjulia_path, ctypes.RTLD_GLOBAL) # <-- avoids segfault
        libjulia.jl_init__threading.argtypes = []
        libjulia.jl_init__threading.restype = None
        libjulia.jl_init__threading()
        libjulia.jl_eval_string.argtypes = [ctypes.c_char_p]
        libjulia.jl_eval_string.restype = ctypes.c_void_p
    finally:
        os.chdir(current_dir)

    return libjulia


the_libjulia = load_libjulia()


def eval_str(code="", libjulia=the_libjulia):
    return libjulia.jl_eval_string(code.encode('utf8'))

def runcom():
    eval_str("import Pkg; Pkg.activate(\"../venv/julia_project/myjuliamod-1.7.2/\"); Pkg.instantiate()")



import os
import sys
import shutil
import sysconfig
import subprocess

import logging
LOGGER = logging.getLogger('julia_project.utils') # shorten this?

is_windows = os.name == "nt"
is_apple = sys.platform == "darwin"

SHLIB_SUFFIX = sysconfig.get_config_var("SHLIB_SUFFIX")
if SHLIB_SUFFIX is None:
    if is_windows:
        SHLIB_SUFFIX = ".dll"
    else:
        SHLIB_SUFFIX = ".so"
if is_apple:
    # sysconfig.get_config_var("SHLIB_SUFFIX") can be ".so" in macOS.
    # Let's not use the value from sysconfig.
    SHLIB_SUFFIX = ".dylib"


# Copied from jill.py
def query_yes_no(question, default="yes"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError(f"invalid default answer: '{default}'")

    while True:
        sys.stdout.write(question + prompt)
        try:
            choice = input().lower()
        except KeyboardInterrupt as err:
            print(f"Got {err}")
            return None
        if default is not None and choice == '':
            return valid[default]
        if choice in valid:
            return valid[choice]
        sys.stdout.write("Please respond with 'yes' or 'no' "
                         "(or 'y' or 'n').\n")


# From PythonCall
def julia_version_str(exe):
    """
    If exe is a julia executable, return its version as a string. Otherwise raise an exception.
    """
    proc = subprocess.run([exe, "--version"], check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    words = proc.stdout.decode('utf-8').split()
    if len(words) < 3 or words[0].lower() != 'julia' or words[1].lower() != 'version':
        raise ValueError(f"{exe} is not a julia executable")
    return words[2]


# Adapted from PythonCall
def default_depot_path():
    return (os.environ.get("JULIA_DEPOT_PATH", "").split(";" if os.name == "nt" else ":")[0]
               or os.path.join(os.path.expanduser("~"), ".julia")
               )


# Adapted from PythonCall
def get_virtual_env_path():
    paths = [os.environ.get('VIRTUAL_ENV'), os.environ.get('CONDA_PREFIX'), os.environ.get('MAMBA_PREFIX')]
    paths = [x for x in paths if x is not None]
    if len(paths) == 0:
        env_path = None
    elif len(paths) == 1:
        env_path = paths[0]
    else:
        raise Exception('You are using some mix of virtual, conda and mamba environments, cannot figure out which to use!')
    return env_path


def maybe_remove(path):
    if os.path.exists(path):
        os.remove(path)
        LOGGER.info(f"Removing {path}")


def update_copy(src, dest):
    """
    Possibly copy `src` to `dest`. No copy unless `src` exists.
    Copy if `dest` does not exist, or mtime of dest is older than
    of `src`.

    Returns: None
    """
    if os.path.exists(src):
        if (not os.path.exists(dest) or
            os.path.getmtime(dest) < os.path.getmtime(src)):
            shutil.copy(src, dest)


def _project_toml(project_path):
    return os.path.join(project_path, "Project.toml")


def _julia_project_toml(project_path):
    return os.path.join(project_path, "JuliaProject.toml")


def no_project_toml_message(project_path):
    return f'Neither "{_project_toml(project_path)}" nor "{_julia_project_toml(project_path)}" exist.'


def has_project_toml(project_path):
    return (
            os.path.exists(_project_toml(project_path)) or
            os.path.exists(_julia_project_toml(project_path))
    )


def has_manifest_toml(_dir):
    return (
            os.path.exists(os.path.join(_dir, "Manifest.toml")) or
            os.path.exists(os.path.join(_dir, "JuliaManifest.toml"))
    )
